@@ -5,9 +5,6 @@ in instead of the defaults.
 */
 
 #include "Rts.h"
-#if defined(__GLASGOW_HASKELL__) && (__GLASGOW_HASKELL__ <611)
-#include "RtsFlags.h"
-#endif
 
 #include "HsFFI.h"
 
@@ -23,7 +20,6 @@ defaultsHook (void)
     RtsFlags.GcFlags.heapSizeSuggestion = 6*1024*1024 / BLOCK_SIZE;
     RtsFlags.GcFlags.maxStkSize         = 512*1024*1024 / sizeof(W_);
     RtsFlags.GcFlags.giveStats = COLLECT_GC_STATS;
-    RtsFlags.GcFlags.statsFile = stderr;
 
     // See #3408: the default idle GC time of 0.3s is too short on
     // Windows where we receive console events once per second or so.
