@@ -208,6 +208,7 @@ module Data.ByteString.Char8 (
         hGet,                   -- :: Handle -> Int -> IO ByteString
         hGetNonBlocking,        -- :: Handle -> Int -> IO ByteString
         hPut,                   -- :: Handle -> ByteString -> IO ()
+        hPutNonBlocking,        -- :: Handle -> ByteString -> IO ByteString
         hPutStr,                -- :: Handle -> ByteString -> IO ()
         hPutStrLn,              -- :: Handle -> ByteString -> IO ()
 
@@ -236,9 +237,9 @@ import Data.ByteString (empty,null,length,tail,init,append
                        ,sort,isPrefixOf,isSuffixOf,isInfixOf
                        ,findSubstring,findSubstrings,breakSubstring,copy,group
 
-                       ,getLine, getContents, putStr, putStrLn, interact
-                       ,hGetContents, hGet, hPut, hPutStr, hPutStrLn
-                       ,hGetLine, hGetNonBlocking
+                       ,getLine, getContents, putStr, interact
+                       ,hGetContents, hGet, hPut, hPutStr
+                       ,hGetLine, hGetNonBlocking, hPutNonBlocking
                        ,packCString,packCStringLen
                        ,useAsCString,useAsCStringLen
                        )
@@ -249,7 +250,7 @@ import Data.ByteString.Internal (ByteString(PS), c2w, w2c, isSpaceWord8
 import Data.Char    ( isSpace )
 import qualified Data.List as List (intersperse)
 
-import System.IO                (openFile,hClose,hFileSize,IOMode(..))
+import System.IO                (Handle,stdout,openFile,hClose,hFileSize,IOMode(..))
 #ifndef __NHC__
 import Control.Exception        (bracket)
 #else
@@ -269,7 +270,7 @@ import GHC.Ptr                  (Ptr(..))
 import GHC.ST                   (ST(..))
 #endif
 
-#if __GLASGOW_HASKELL__ >= 608
+#if MIN_VERSION_base(3,0,0)
 import Data.String              (IsString(..))
 #endif
 
@@ -285,7 +286,7 @@ singleton :: Char -> ByteString
 singleton = B.singleton . c2w
 {-# INLINE singleton #-}
 
-#if __GLASGOW_HASKELL__ >= 608
+#if MIN_VERSION_base(3,0,0)
 instance IsString ByteString where
     fromString = pack
     {-# INLINE fromString #-}
@@ -544,15 +545,12 @@ break f = B.break (f . w2c)
 {-# INLINE [1] break #-}
 #endif
 
-#if __GLASGOW_HASKELL__ >= 606
--- This RULE LHS is not allowed by ghc-6.4
 {-# RULES
 "ByteString specialise break (x==)" forall x.
     break ((==) x) = breakChar x
 "ByteString specialise break (==x)" forall x.
     break (==x) = breakChar x
   #-}
-#endif
 
 -- INTERNAL:
 
@@ -1039,3 +1037,13 @@ appendFile :: FilePath -> ByteString -> IO ()
 appendFile f txt = bracket (openFile f AppendMode) hClose
     (\h -> hPut h txt)
 
+
+-- | Write a ByteString to a handle, appending a newline byte
+hPutStrLn :: Handle -> ByteString -> IO ()
+hPutStrLn h ps
+    | length ps < 1024 = hPut h (ps `B.snoc` 0x0a)
+    | otherwise        = hPut h ps >> hPut h (B.singleton (0x0a)) -- don't copy
+
+-- | Write a ByteString to stdout, appending a newline byte
+putStrLn :: ByteString -> IO ()
+putStrLn = hPutStrLn stdout
