@@ -112,6 +112,11 @@ import GHC.IO.Exception	( IOException(..), IOErrorType(..), ioException )
 import GHC.IOBase	( IOException(..), IOErrorType(..), ioException )
 #endif
 
+#if __GLASGOW_HASKELL__ > 700
+import GHC.IO.Encoding
+import GHC.Foreign as GHC
+#endif
+
 #ifdef mingw32_HOST_OS
 import System.Posix.Types
 import System.Posix.Internals
@@ -709,11 +714,18 @@ canonicalizePath :: FilePath -> IO FilePath
 canonicalizePath fpath =
 #if defined(mingw32_HOST_OS)
     do path <- Win32.getFullPathName fpath
+#else
+#if __GLASGOW_HASKELL__ > 700
+  GHC.withCString fileSystemEncoding fpath $ \pInPath ->
+  allocaBytes long_path_size $ \pOutPath ->
+    do throwErrnoPathIfNull "canonicalizePath" fpath $ c_realpath pInPath pOutPath
+       path <- GHC.peekCString fileSystemEncoding pOutPath
 #else
   withCString fpath $ \pInPath ->
   allocaBytes long_path_size $ \pOutPath ->
     do throwErrnoPathIfNull "canonicalizePath" fpath $ c_realpath pInPath pOutPath
        path <- peekCString pOutPath
+#endif
 #endif
        return (normalise path)
         -- normalise does more stuff, like upper-casing the drive letter
@@ -862,6 +874,11 @@ Insufficient resources are available to perform the operation.
 * 'UnsupportedOperation'
 The operating system has no notion of current directory.
 
+Note that in a concurrent program, the current directory is global
+state shared between all threads of the process.  When using
+filesystem operations from multiple threads, it is therefore highly
+recommended to use absolute rather than relative `FilePath`s.
+
 -}
 #ifdef __GLASGOW_HASKELL__
 getCurrentDirectory :: IO FilePath
@@ -902,6 +919,11 @@ current directory cannot be dynamically changed.
 The path refers to an existing non-directory object.
 @[ENOTDIR]@
 
+Note that in a concurrent program, the current directory is global
+state shared between all threads of the process.  When using
+filesystem operations from multiple threads, it is therefore highly
+recommended to use absolute rather than relative `FilePath`s.
+
 -}
 
 setCurrentDirectory :: FilePath -> IO ()
