@@ -24,8 +24,8 @@
  * modules these names are defined in.
  */
 
-PRELUDE_CLOSURE(ghczmprim_GHCziBool_True_closure);
-PRELUDE_CLOSURE(ghczmprim_GHCziBool_False_closure);
+PRELUDE_CLOSURE(ghczmprim_GHCziTypes_True_closure);
+PRELUDE_CLOSURE(ghczmprim_GHCziTypes_False_closure);
 PRELUDE_CLOSURE(base_GHCziPack_unpackCString_closure);
 PRELUDE_CLOSURE(base_GHCziWeak_runFinalizzerBatch_closure);
 
@@ -82,8 +82,8 @@ PRELUDE_INFO(base_GHCziWord_W64zh_con_info);
 PRELUDE_INFO(base_GHCziStable_StablePtr_static_info);
 PRELUDE_INFO(base_GHCziStable_StablePtr_con_info);
 
-#define True_closure              DLL_IMPORT_DATA_REF(ghczmprim_GHCziBool_True_closure)
-#define False_closure             DLL_IMPORT_DATA_REF(ghczmprim_GHCziBool_False_closure)
+#define True_closure              DLL_IMPORT_DATA_REF(ghczmprim_GHCziTypes_True_closure)
+#define False_closure             DLL_IMPORT_DATA_REF(ghczmprim_GHCziTypes_False_closure)
 #define unpackCString_closure     DLL_IMPORT_DATA_REF(base_GHCziPack_unpackCString_closure)
 #define runFinalizerBatch_closure DLL_IMPORT_DATA_REF(base_GHCziWeak_runFinalizzerBatch_closure)
 #define mainIO_closure            (&ZCMain_main_closure)
