@@ -9,12 +9,13 @@
 #ifndef PROFHEAP_H
 #define PROFHEAP_H
 
+#include "GetTime.h" // for Ticks
+
 #include "BeginPrivate.h"
 
-void    heapCensus         (void);
+void    heapCensus         (Ticks t);
 nat     initHeapProfiling  (void);
 void    endHeapProfiling   (void);
-void    LDV_recordDead     (StgClosure *c, nat size);
 rtsBool strMatchesSelector (char* str, char* sel);
 
 #include "EndPrivate.h"
